;;;; qam.lisp
;;;;
;;;; Author: Robert Smith

(in-package #:qvm)

(defgeneric run (qam)
  (:documentation "Simulate the quantum abstract machine QAM until completion. Return the QAM in its end state."))

(defgeneric reset-quantum-state (qam)
  (:documentation "Bring all qubits of the quantum abstract machine QAM to the zero state."))

(defgeneric measure (qam q)
  (:documentation  "Non-deterministically perform a measurement on the qubit addressed by Q in the quantum abstract machine QAM.

Return two values:

    1. The resulting QAM.
    2. The measured classical bit."))

(defun measure-and-store (qam q c)
  "Performs a measurement on the qubit addressed by Q in the quantum abstract machine QAM, and stores the measured
bit in the classical bit addressed by C.

Return two values:

    1. The resulting QAM.
    2. The measured classical bit."
  (check-type c quil:memory-ref)
  (multiple-value-bind (ret-qam cbit)
      (measure qam q)
    (setf (dereference-mref qam c) cbit)
    (values ret-qam cbit)))

(defgeneric measure-all (qam)
  (:documentation "Non-deterministically perform a measurement on all qubits in the quantum abstract machine QAM.
Return two values:

    1. The resulting QAM.
    2. A list of measured bits.")
  (:method ((qam t))
    (let ((measured-bits nil))
      (loop :for q :from (1- (number-of-qubits qam)) :downto 0
            :do (multiple-value-bind (ret-qam bit)
                    (measure qam q)
                  (push bit measured-bits)
                  (setf qam ret-qam)))
      (values
       qam
       measured-bits))))

(defgeneric number-of-qubits (qam)
  (:documentation "Return the number of qubits configured on the quantum abstract machine QAM."))

(defgeneric compile-loaded-program (qam)
  (:documentation "Compile the program loaded into the qam QAM so as to optimize execution."))
